

def elim_apostrophes(args):
    """ strip apostrphes form items in args, allows for user to enslose characters and phrases that cannot naturally
    be parsed, such as spaces, or items beginning in a hyphen like negative numbers"""
    for key in args.keys():
        if isinstance(args[key], list):
            for i in range(0, len(args[key])):
                args[key][i] = args[key][i].replace("'", '')
        elif isinstance(args[key], str):
            args[key] = args[key].replace("'", '')


def process_clause(args):
    w_params = []
    for cond in args['<w_conds>']:
        if cond == 'eq':
            w_params.append('=')
        if cond == 'gt':
            w_params.append('>=')
        if cond == 'lt':
            w_params.append('<=')
        if cond == 'like':
            w_params.append('LIKE')

    return w_params


def o_cond(args):
    o_cond = ""
    if args['--asc']:
        o_cond = 'ASC'
    elif args['--desc']:
        o_cond = 'DESC'

    return o_cond


import re
import dateutil
from datetime import datetime

import pandas as pd
from textblob.classifiers import NaiveBayesClassifier
from colorama import init, Fore, Style
from tabulate import tabulate

import finance_db


class BankClassify:

    def __init__(self, data="sql"):
        """Load in the previous data (by default from `data`) and initialise the classifier"""
        self.db = finance_db.FinanceDB()

        # allows dynamic training data to be used (i.e many accounts in a loop)
        self.unprocessed_data = None
        self.processed_data = None
        self.categories = None

        if data == 'sql':

            # select all data from transactions joined with categories descriptions that have been assigned a
            # category to be used for/as training data - only data with a category is selected
            if 'categories'not in self.db.schema.keys():
                self.db.create_table('categories')
                self.db.conn.commit()

            data = self.db.conn.cursor().execute("SELECT date, desc, amount, cat_desc FROM transactions "
                                                 "JOIN categories ON transactions.cat_id=categories.cat_id").fetchall()
            # create data frame from cursor selection with expected column names
            prev_df = pd.DataFrame(data, columns=['date', 'desc', 'amount', 'cat'])

            self.training_data = prev_df
        else:
            self.training_data = pd.DataFrame(columns=['date', 'desc', 'amount', 'cat'])

        self.classifier = NaiveBayesClassifier(self._get_training(self.training_data), self._extractor)

    def add_unprocessed_data(self):
        """Add new data and interactively classify it.
        Arguments:
         - filename: filename of Santander-format file
        """
        query = pd.read_sql_query("SELECT * FROM transactions "
                                  "WHERE cat_id IS NULL ORDER BY date asc", self.db.conn)
        unprocessed_df = pd.DataFrame(query, columns=self.db.schema['transactions'])

        self.unprocessed_data = unprocessed_df

        # update following section to add data back into transactions db

        # self.prev_data = pd.concat([self.prev_data, self.new_data])
        # # save data to the same file we loaded earlier
        # self.prev_data.to_csv(self.trainingDataFile, index=False)

    def _prep_for_analysis(self):
        """Prepare data for analysis in pandas, setting index types and subsetting"""
        self.training_data = self._make_date_index(self.training_data)

        self.training_data['cat'] = self.training_data['cat'].str.strip()

        self.inc = self.training_data[self.training_data.amount > 0]
        self.out = self.training_data[self.training_data.amount < 0]
        self.out.amount = self.out.amount.abs()

        self.inc_noignore = self.inc[self.inc.cat != 'Ignore']
        self.inc_noexpignore = self.inc[(self.inc.cat != 'Ignore') & (self.inc.cat != 'Expenses')]

        self.out_noignore = self.out[self.out.cat != 'Ignore']
        self.out_noexpignore = self.out[(self.out.cat != 'Ignore') & (self.out.cat != 'Expenses')]

    def _read_categories(self):
        """Read list of categories from categories.txt"""
        categories = {}

        with open('categories.txt') as f:
            for i, line in enumerate(f.readlines()):
                categories[i] = line.strip()

        return categories

    def _read_categories_sql(self):
        """Read list of categories from categories.txt
            categories in an enumerated dictionary!
            """
        categories_dict = {}
        categories = self.db.conn.cursor().execute("SELECT * FROM categories").fetchall()

        categories_dict = dict((x, y) for x, y in categories)

        return categories_dict

    def _add_new_category(self, categories, category):
        """Add a new category to categories.txt"""
        # with open('categories.txt', 'a') as f:
        #     f.write('\n' + category)
        if len(categories) != 0:
            max_key = max(list(categories.keys()))
            new_key = max_key + 1
            categories[new_key] = category
        else:
            categories[1] = category

        return categories

    def commit_categories(self, table='categories'):
        # move categories dictionary to dataframe to make sql transition easier
        if self.categories:
            categories_df = pd.DataFrame.from_dict(data=self.categories, orient="index")
            # replace because entire category is added when loaded, ignore is deal but not available
            categories_df.rename(columns={0: self.db.schema['categories'][1]}, inplace=True)
            categories_df.index.names = [self.db.schema['categories'][0],]
            # TODO cant replace table when there are foreign key constraints
            # TODO compare old and new categories OUTSIDE of database, then reindex and append without index to allow for autoincrement
            self.db.drop_duplicates(table='categories', method='outside', new_data=categories_df)

            # categories_df.to_sql(name='categories', con=self.db.conn, index=False, if_exists='append')

            self.db.conn.commit()

    def commit_data(self, table):
        if not self.processed_data.empty:
            # set processed column to 1 for processed columns
            self.processed_data['processed'] = 1
            self.processed_data.drop(columns=['trans_id'], inplace=True)
            self.processed_data.to_sql(name=table, con=self.db.conn, index=False, if_exists='append')

            self.db.drop_duplicates(table=table, condition='MAX', method='inside',
                                    filter_columns=['date', 'desc', 'amount', 'total_id'])
            self.db.conn.commit()

    # function to return key for any value
    def get_key(self, val):
        for key, value in self.categories.items():
            if val == value:
                return key

        return "key doesn't exist"

    def ask_with_guess(self):
        """Interactively guess categories for each transaction in df, asking each time if the guess
        is correct"""
        # Initialise colorama
        init()

        # new_data['cat'] = ""
        self.add_unprocessed_data()
        self.unprocessed_data.rename(columns={"description": "desc"}, inplace=True)
        self.processed_data = pd.DataFrame(columns=self.db.schema['transactions'])

        # categories returned as dictionary
        self.categories = self._read_categories_sql()
        if not self.categories:
            self.categories = dict()

        for index, row in self.unprocessed_data.iterrows():
            # Generate the category numbers table from the list of categories
            cats_list = [[idnum, cat] for idnum, cat in self.categories.items()]
            cats_table = tabulate(cats_list)

            stripped_text = self._strip_numbers(row['desc'])

            # Guess a category using the classifier (only if there is data in the classifier)
            if len(self.classifier.train_set) > 1:
                guess = self.classifier.classify(stripped_text)
            else:
                guess = ""

            # Print list of categories
            print(cats_table)
            print("\n")
            # Print transaction
            # print("On: %s\t %.2f\n%s" % (row['date'], row['amount'], row['desc']))
            # print(Fore.RED  + Style.BRIGHT + "My guess is: " + str(guess) + Fore.RESET)
            print(tabulate(self.unprocessed_data.loc[[index]], headers='keys'))
            print(Fore.RED + Style.BRIGHT + "My guess is: " + str(guess) + Fore.RESET)

            input_value = input("> ")
            category = ""
            if input_value.lower() == 'q':
                # If the input was 'q' then quit
                self.commit_categories()
                self.commit_data(table='transactions')
                return

            if input_value == "":
                # If the input was blank then our guess was right!
                # TODO need to put in category index not description or sql will be pissed
                # update data with category ID
                self.unprocessed_data.at[index, 'cat_id'] = self.get_key(guess)
                category_key = self.get_key(guess)
                # update classifier with category (guess)
                self.classifier.update([(stripped_text, guess)])

            else:
                # Otherwise, our guess was wrong
                if input_value in self.categories.values():
                    # if entered category is a categories dictionary value, convert ot integer id and update data
                    category = input_value
                    category_key = self.get_key(input_value)
                else:
                    try:
                        category_key = int(input_value)
                        if int(input_value) in self.categories.keys():
                            # if integer key entered for category use as is
                            category_key = input_value
                            category = self.categories[int(input_value)]

                    except ValueError:
                        # Otherwise, we've entered a new category, so add it to the list of
                        # categories
                        category = input_value
                        self._add_new_category(self.categories, input_value)
                        category_key = self.get_key(input_value)

            # Write correct answer
            self.unprocessed_data.at[index, 'cat_id'] = category_key
            self.processed_data = self.processed_data.append(self.unprocessed_data.loc[[index]])
            # Update classifier
            self.classifier.update([(stripped_text, category)])

            # if all transactions are processed
        self.commit_categories()
        self.commit_data(table='transactions')

        return

    def _make_date_index(self, df):
        """Make the index of df a Datetime index"""
        df.index = pd.DatetimeIndex(df.date.apply(dateutil.parser.parse, dayfirst=True))

        return df

    def _read_nationwide_file(self, filename):
        """Read a file in the csv file that Nationwide provides downloads in.
        Returns a pd.DataFrame with columns of 'date', 'desc' and 'amount'."""

        with open(filename) as f:
            lines = f.readlines()


        dates = []
        descs = []
        amounts = []

        for line in lines[5:]:

            line = "".join(i for i in line if ord(i)<128)
            if line.strip() == '':
                continue

            splits = line.split("\",\"")
            """
            0 = Date
            1 = Transaction type
            2 = Description
            3 = Paid Out
            4 = Paid In
            5 = Balance
            """
            date = splits[0].replace("\"", "").strip()
            date = datetime.strptime(date, '%d %b %Y').strftime('%d/%m/%Y')
            dates.append(date)

            # get spend/pay in amount
            if splits[3] != "": # paid out
                spend = float(re.sub("[^0-9\.-]", "", splits[3])) * -1
            else: # paid in
                spend = float(re.sub("[^0-9\.-]", "", splits[4]))

            amounts.append(spend)

            #Description
            descs.append(splits[2])

        df = pd.DataFrame({'date':dates, 'desc':descs, 'amount':amounts})

        df['amount'] = df.amount.astype(float)
        df['desc'] = df.desc.astype(str)
        df['date'] = df.date.astype(str)

        return df

    def _read_santander_file(self, filename):
        """Read a file in the plain text format that Santander provides downloads in.
        Returns a pd.DataFrame with columns of 'date', 'desc' and 'amount'."""
        with open(filename, errors='replace') as f:
            lines = f.readlines()

        dates = []
        descs = []
        amounts = []

        for line in lines[4:]:

            line = "".join(i for i in line if ord(i)<128)
            if line.strip() == '':
                continue

            splitted = line.split(":")

            category = splitted[0]
            data = ":".join(splitted[1:])

            if category == 'Date':
                dates.append(data.strip())
            elif category == 'Description':
                descs.append(data.strip())
            elif category == 'Amount':
                just_numbers = re.sub("[^0-9\.-]", "", data)
                amounts.append(just_numbers.strip())


        df = pd.DataFrame({'date':dates, 'desc':descs, 'amount':amounts})


        df['amount'] = df.amount.astype(float)
        df['desc'] = df.desc.astype(str)
        df['date'] = df.date.astype(str)

        return df

    def _read_lloyds_csv(self, filename):
        """Read a file in the CSV format that Lloyds Bank provides downloads in.
        Returns a pd.DataFrame with columns of 'date' 0 , 'desc'  4 and 'amount' 5 ."""

        df = pd.read_csv(filename, skiprows=0)

        """Rename columns """
        #df.columns = ['date', 'desc', 'amount']
        df.rename(
            columns={
                "Transaction Date" : 'date',
                "Transaction Description" : 'desc',
                "Debit Amount": 'amount',
                "Credit Amount": 'creditAmount'
            },
            inplace=True
        )

        # if its income we still want it in the amount col!
        # manually correct each using 2 cols to create 1 col with either + or - figure
        # lloyds outputs 2 cols, credit and debit, we want 1 col representing a +- figure
        for index, row in df.iterrows():
            if (row['amount'] > 0):
                # it's a negative amount because this is a spend
                df.at[index, 'amount'] = -row['amount']
            elif (row['creditAmount'] > 0):
                df.at[index, 'amount'] = row['creditAmount']

        # cast types to columns for math
        df = df.astype({"desc": str, "date": str, "amount": float})

        return df

    def _read_mint_csv(self, filename) -> pd.DataFrame:
        """Read a file in the CSV format that mint.intuit.com provides downloads in.
        Returns a pd.DataFrame with columns of 'date', 'desc', and 'amount'."""

        df = pd.read_csv(filename, skiprows=0)

        """Rename columns """
        # df.columns = ['date', 'desc', 'amount']
        df.rename(
            columns={
                "Date": 'date',
                "Original Description": 'desc',
                "Amount": 'amount',
                "Transaction Type": 'type'
            },
            inplace=True
        )

        # mint outputs 2 cols, amount and type, we want 1 col representing a +- figure
        # manually correct amount based on transaction type colum with either + or - figure
        df.loc[df['type'] == 'debit', 'amount'] = -df['amount']

        # cast types to columns for math
        df = df.astype({"desc": str, "date": str, "amount": float})
        df = df[['date', 'desc', 'amount']]

        return df

    def _read_barclays_csv(self, filename):
        """Read a file in the CSV format that Barclays Bank provides downloads in.
        Edge case: foreign txn's sometimes causes more cols than it should
        Returns a pd.DataFrame with columns of 'date' 1 , 'desc' (memo)  5 and 'amount' 3 ."""

        # Edge case: Barclays foreign transaction memo sometimes contains a comma, which is bad.
        # Use a work-around to read only fixed col count
        # https://stackoverflow.com/questions/20154303/pandas-read-csv-expects-wrong-number-of-columns-with-ragged-csv-file
        # Prevents an error where some rows have more cols than they should
        temp=pd.read_csv(filename,sep='^',header=None,prefix='X',skiprows=1)
        temp2=temp.X0.str.split(',',expand=True)
        del temp['X0']
        df = pd.concat([temp,temp2],axis=1)

        """Rename columns """
        df.rename(
            columns={
                1: 'date',
                5 : 'desc',
                3: 'amount'
            },
            inplace=True
        )

        # cast types to columns for math
        df = df.astype({"desc": str, "date": str, "amount": float})

        return df

    def _get_training(self, df):
        """Get training data for the classifier, consisting of tuples of
        (text, category)"""
        train = []
        subset = df[df['cat'] != '']
        for i in subset.index:
            row = subset.iloc[i]
            new_desc = self._strip_numbers(row['desc'])
            train.append((new_desc, row['cat']))

        return train

    def _extractor(self, doc):
        """Extract tokens from a given string"""
        # TODO: Extend to extract words within words
        # For example, MUSICROOM should give MUSIC and ROOM
        tokens = self._split_by_multiple_delims(doc, [' ', '/'])

        features = {}

        for token in tokens:
            if token == "":
                continue
            features[token] = True

        return features

    def _strip_numbers(self, s):
        """Strip numbers from the given string"""
        return re.sub("[^A-Z ]", "", s)

    def _split_by_multiple_delims(self, string, delims):
        """Split the given string by the list of delimiters given"""
        regexp = "|".join(delims)

        return re.split(regexp, string)


from web3 import Web3
import json
from decimal import Decimal

import pricing

w3 = Web3(Web3.HTTPProvider('https://mainnet.infura.io/v3/80de2f7c4deb45d5a880a822f2bb1e5d'))
# define ERC20 contract
ERC20_ABI = json.loads('[{"constant":true,"inputs":[],"name":"name","outputs":[{"name":"","type":"string"}],'
                       '"payable":false,"stateMutability":"view","type":"function"},{"constant":false,'
                       '"inputs":[{"name":"_spender","type":"address"},{"name":"_value","type":"uint256"}],'
                       '"name":"approve","outputs":[{"name":"","type":"bool"}],"payable":false,'
                       '"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],'
                       '"name":"totalSupply","outputs":[{"name":"","type":"uint256"}],"payable":false,'
                       '"stateMutability":"view","type":"function"},{"constant":false,'
                       '"inputs":[{"name":"_from","type":"address"},{"name":"_to","type":"address"},{"name":"_value",'
                       '"type":"uint256"}],"name":"transferFrom","outputs":[{"name":"","type":"bool"}],'
                       '"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,"inputs":[],'
                       '"name":"decimals","outputs":[{"name":"","type":"uint8"}],"payable":false,'
                       '"stateMutability":"view","type":"function"},{"constant":true,'
                       '"inputs":[{"name":"_owner","type":"address"}],"name":"balanceOf",'
                       '"outputs":[{"name":"","type":"uint256"}],"payable":false,"stateMutability":"view",'
                       '"type":"function"},{"constant":true,"inputs":[],"name":"symbol",'
                       '"outputs":[{"name":"","type":"string"}],"payable":false,"stateMutability":"view",'
                       '"type":"function"},{"constant":false,"inputs":[{"name":"_to","type":"address"},'
                       '{"name":"_value","type":"uint256"}],"name":"transfer","outputs":[{"name":"","type":"bool"}],'
                       '"payable":false,"stateMutability":"nonpayable","type":"function"},{"constant":true,'
                       '"inputs":[{"name":"_owner","type":"address"},{"name":"_spender","type":"address"}],'
                       '"name":"allowance","outputs":[{"name":"","type":"uint256"}],"payable":false,'
                       '"stateMutability":"view","type":"function"},{"anonymous":false,'
                       '"inputs":[{"indexed":true,"name":"_from","type":"address"},'
                       '{"indexed":true,"name":"_to","type":"address"},'
                       '{"indexed":false,"name":"_value","type":"uint256"}],"name":"Transfer","type":"event"},'
                       '{"anonymous":false,"inputs":[{"indexed":true,"name":"_owner","type":"address"},'
                       '{"indexed":true,"name":"_spender","type":"address"},'
                       '{"indexed":false,"name":"_value","type":"uint256"}],"name":"Approval","type":"event"}]')


def update_holdings(db):
    if not db.exists('crypto'):
        db.create_table('crypto')

    # get wallet address from db
    wallet_addy = db.conn.cursor().execute("SELECT num FROM accounts WHERE institution = 'crypto'").fetchall()[0][0]

    # get holdings from database
    holdings = db.conn.cursor().execute("SELECT * FROM crypto_holdings").fetchall()

    # create dictionary with holdings table schema as keys
    holding = dict()
    # holding = {key: None for key in db.schema['crypto_holdings']}

    for entry in holdings:
        for i in range(0, len(entry)):
            holding[db.schema['crypto_holdings'][i]] = entry[i]

        update = {key: None for key in db.schema['crypto']}
        update['desc'] = holding['desc']

        # get token quantity using native decimal count ether is special case
        if holding['desc'] == 'ether':
            wei_balance = w3.eth.get_balance(wallet_addy)
            eth_balance = w3.fromWei(wei_balance, 'ether')
            update['qty'] = float(eth_balance)
        else:
            token = w3.eth.contract(address=holding['chain_address'], abi=ERC20_ABI)
            token_dec = token.functions.decimals().call()
            token_balance = token.functions.balanceOf(wallet_addy).call()
            token_balance = token_balance/Decimal(10 ** token_dec)

            update['qty'] = float(token_balance)

        update['price'] = pricing.get_price(holding['symbol'])
        update['total'] = update['qty'] * update['price']

        update_filtered = {k: v for k, v in update.items() if v is not None}

        update_cols = list(update_filtered.keys())
        update_vals = (list(update_filtered.values()))

        query = db.insert(table='crypto', columns=update_cols, values=update_vals)
        db.conn.cursor().execute(query, update_vals)
        db.conn.commit()


import pandas_datareader as datar


def get_price(symbol, source='yahoo'):
    """ give a symobl -> get a price """

    price = datar.DataReader(symbol, source)
    price = price["Close"][price.last_valid_index()]
    return price



from pathlib import Path
from web3 import Web3

import pandas as pd
import os
import sqlite3
import tabulate
import copy

# dans libraries
import crypto
import finance_db
import questrade_2 as qt
from classify import BankClassify

fipy_fp = Path(__file__).absolute().parent


def get_header(db, table, columns=None):
    """ gets proper header for data if tabulated given columns as string

    args:
        db:             db object from finance_db
        table:          string table name to query
        columns:        select columns as list of strings toi be filtered in query
    """

    header = []
    if not columns or '*' in columns:
        header = db.schema[table]
    else:
        for column in columns:
            if column in db.schema[table]:
                header.append(column)

    return header


def show_tabulated_sql(db, query, data=None):
    """ pass db object and query object to show tabulated sql data

    args:
        db:             db object from finance_db
        query:          query object from graph.py
        data:           iterable data as list of tuples or dataframe

    """
    # passed query is used to get data if no data is passed
    if not data:
        data = db.conn.cursor().execute(query.build_select()).fetchall()

    header = get_header(db, query.table, query.s_cols)
    print("")
    print(tabulate.tabulate(data, headers=header))
    print("")


def create_acc(query):
    """ Modify query object per the requirements of account creation
    Args:
        query:          Query object from finance_db library

    """

    # assign file path to accounts file within source parent
    filepath = None
    number = None
    for i in range(0, len(query.in_cols)):
        if query.in_cols[i] == 'num':
            filepath = str(fipy_fp.joinpath('accounts').joinpath(query.in_vals[i]))
            number = query.in_vals[i]

    for i in range(0, len(query.in_cols)):
        if query.in_cols[i] == 'filepath':
            query.in_vals[i] = filepath
            query.build_str()

    source = None
    for i in range(0, len(query.in_cols)):
        if query.in_cols[i] == 'source':
            source = query.in_vals[i]

    # create directory for saving transaction data to be imported if file is source
    if source == 'file':
        os.makedirs(fipy_fp.joinpath('accounts'), exist_ok=True)
        os.makedirs(fipy_fp.joinpath('accounts').joinpath(number), exist_ok=True)


def create_holding(query):
    # convert chain address to checksum before storage
    for i in range(0, len(query.in_cols)):
        if query.in_cols[i] == 'chain_address':
            query.in_vals[i] = Web3.toChecksumAddress(query.in_vals[i])
            query.build_str()

    return


def create_budget(db, query):
    # create tags table if none exists
    if not db.exists('budgets'):
        db.create_table('budgets')

    db.conn.cursor().execute(query.build_insert())

    # delete duplicates and keep minimum id value (original entry)
    db.drop_duplicates(table=query.table, condition='MIN', method='inside', filter_columns=query.s_cols)
    db.conn.commit()

    show_tabulated_sql(db, query)

    print("Created budget successfully")

    return


def create_dashboard(db, query):
    # create tags table if none exists
    if not db.exists('dashboards'):
        db.create_table('dashboards')

    db.conn.cursor().execute(query.build_insert())

    # delete duplicates and keep minimum id value (original entry)
    db.drop_duplicates(table=query.table, condition='MIN', method='inside', filter_columns=query.s_cols)
    db.conn.commit()

    print("Created dashboard successfully")

    return


def create_item(db, query, item_type, drop_cond=None, drop_method=None, comp_columns=None):
    # create table if none exists to put items in
    if not db.exists(query.table):
        db.create_table(query.table)

    if item_type == 'account':
        create_acc(query=query)
    elif item_type == 'crypto_holding':
        create_holding(query=query)

    db.conn.cursor().execute(query.build_insert())

    if drop_cond and drop_method:
        # delete duplicates and keep minimum id value (original entry)
        db.drop_duplicates(table=query.table, condition=drop_cond, method=drop_method, filter_columns=comp_columns)
        db.conn.commit()

    # show new table entry
    show_tabulated_sql(db, query)

    print(item_type + " created successfully!")


def td_csv2df(db, new_data, filepath, acc_id):
    """ converts ALL TD type csv files in the account folder to appropriate dataframe for importing to sql and
        interfacing with other functions

        Args:
        account         account number or description as string
        new_data        blank dataframe with columns [date, description, amount , account id]
        db              FinanceDB object
    """

    # for each file dumped in the account folder, csv is read and appended to data
    for file in filepath:
        # if file.suffix == ".csv":
        td_statement = file

        # read from csv to df, move withdrawl column into deposit columns and reverse sign, drop withdrawl column
        statement_df = pd.read_csv(td_statement, header=None)
        statement_df[3].fillna(statement_df[2]*-1, inplace=True)

        # drop withdrawls column and total column from csv
        statement_df.drop(columns=[2], inplace=True)
        statement_df.columns = ['date', 'desc', 'amount', 'total_id']
        statement_df['acc_id'] = acc_id
        statement_df['date'] = pd.to_datetime(statement_df['date'])  # format dates and datetime objects
        if new_data.empty:
            new_data = statement_df
        else:

            new_data = pd.concat([new_data, statement_df], ignore_index=True, )
            new_data.drop_duplicates(inplace=True)

    return new_data


def update_account(db, account):
    # create transactions table if no exists
    if not db.exists('transactions'):
        db.create_table('transactions')

    try:
        # get list of account info: acc_id, num, institution, desc, filepath, source if account identifier input can
        # be matched to either the account number or description
        account_info = db.conn.cursor().execute("SELECT * FROM accounts WHERE acc_id=? OR num=? OR desc=?",
                                                (account, account, account)).fetchall()
        # define account information to be used later more discretely
        acc_filepath = Path(account_info[0][4]).rglob('*.csv')
        acc_source = account_info[0][5]
        acc_id = account_info[0][0]
        acc_insti = account_info[0][2]

    except sqlite3.OperationalError:
        print("Could not find Account")
        return

    # if the account source is file
    if acc_source == 'file':
        new_data = pd.DataFrame(columns=['date', 'desc', 'amount', 'acc_id', 'total'])  # date, desc, amount
        new_data['date'] = pd.to_datetime(new_data['date'])  # convert date column to datetime object
        # if account institution is TD, use TD function to convert expected csv
        if acc_insti == 'TD':
            table = 'transactions'
            new_data = td_csv2df(db, new_data, acc_filepath, acc_id)
            new_data.to_sql(name=table, index=False, con=db.conn, if_exists='append')
            db.drop_duplicates(table=table, condition='MIN', method='inside', filter_columns=['date', 'desc', 'amount', 'total_id'])

            # delete any transactions in the splits table form transactions to prevent duplication
            db.conn.cursor().execute("DELETE FROM transactions WHERE trans_id IN "
                                     "(SELECT transactions.trans_id FROM transactions "
                                     "LEFT JOIN splits ON splits.total_id = transactions.total_id "
                                     "WHERE splits.date = transactions.date "
                                     "AND splits.desc = transactions.desc "
                                     "AND splits.amount = transactions.amount "
                                     "AND splits.total_id = transactions.total_id)")

    elif acc_source == 'api':
        if acc_insti == 'QT':
            account_id = db.conn.cursor().execute("SELECT num FROM accounts WHERE institution='QT'").fetchall()[0][0]
            qt.update_qpositions(db=db, account_id=account_id)
        if acc_insti == 'crypto':
            crypto.update_holdings(db)

    db.conn.commit()


def view(db, query):

    """ process requests to view db data - not great right now
        Args:
        table:          table name as string
        columns:        list of column names as string to be displayed DEFAULT: "*" (all columns)
        where_cols:     list of columns to compare to where_vals

        pass options in *args
        pass args ofr options in **kwargs

    """

    data = pd.read_sql(query.build_select(), db.conn)
    if 'amount' in data.columns:
        data.loc['Total', 'amount'] = data['amount'].sum()

    header = get_header(db, query.table, query.s_cols)
    print(tabulate.tabulate(data, headers=header))


def split_transaction(db, query, percentage=50, amount=None):
    """ Split a transaction into multiple and edit the resulting transaction
    Args:
        db:                 FinanceDB object from finance_db lib
        query:              Query object from finance_db lib
        percentage:         percentage by which to split transaction amount, split amount is existing * percentage
                            default = 50
        amount:             amount of split portion of transaction as str, if none defaults to percentage
        """

    # get existing transaction - can only split one transaction at a time
    existing_entry = db.conn.cursor().execute(query.build_select()).fetchall()[0]

    # map existing entry to dictionary with columns as keys for easy handling
    existing_dict = dict()
    for i in range(0, len(existing_entry)):
        existing_dict[db.schema[query.table][i]] = existing_entry[i]

    # create matching new_dict to set up inserting new values
    new_dict = copy.deepcopy(existing_dict)
    new_query = copy.deepcopy(query)

    # insert existing transaction into splits table
    # splits_query = copy.deepcopy(query) # use same query but swap table
    query.table = 'splits'
    query.in_cols = ['trans_id', 'date', 'desc', 'amount', 'total_id']
    query.in_vals = []
    for column in query.in_cols:
        query.in_vals.append(str(existing_dict[column]))

    query.build_str()
    db.conn.cursor().execute(query.build_insert())

    # get updated amount for split transaction
    new_amount = 0
    if amount:
        if not isinstance(amount, int):
            amount = float(amount)

        if new_amount <= abs(existing_dict['amount']):
            # set updated amounts and totals (total_id)
            # new value entered is value of new transaction created by split
            new_amount = amount
        else:
            print('New split transaction amount cannot exceed original')

    elif percentage:
        if not isinstance(percentage, int):
            percentage = int(percentage)
        if int(percentage) >= 1:
            percentage = percentage/100

        # percentage is applied to total to get new amount
        new_amount = existing_dict['amount']*percentage

    # transactions are only recorded in 2 decimal places in typical transaction
    new_amount = round(new_amount, 2)

    # set new amounts for both new and existing transactions
    # set total_id for existing and split transacitons - total_id + amount = total_id prior to transaction
    original_total_id = existing_dict['total_id'] + existing_dict['amount']
    existing_dict['amount'] = existing_dict['amount'] - new_amount
    new_dict['amount'] = new_amount

    existing_dict['total_id'] = original_total_id + existing_dict['amount']
    # adding the split amount should bring the total back up to the original
    new_dict['total_id'] = original_total_id

    # edit existing transaction
    # important to note updates do not apply to the existing transaction - therefore up_cols is initialized empty
    query.up_cols = []
    query.up_cols.append('amount')
    query.up_cols.append('total_id')

    # important to note updates do not apply to the existing transaction - therefore up_vals is initialized empty
    query.up_vals = []
    query.up_vals.append(str(existing_dict['amount']))
    query.up_vals.append(str(existing_dict['total_id']))

    query.build_str()

    edit(db=db, query=query)

    # insert new split transaction with edits
    # insert updates into new_dict
    if new_query.up_cols and new_query.up_vals:
        for i in range(0, len(new_query.up_cols)):
            new_dict[new_query.up_cols[i]] = new_query.up_vals[i]

    # remove original trans_id so new is auto generated
    new_dict['trans_id'] = 'NULL'

    # map new_dict to new_query insert properties
    new_query.in_cols = []
    new_query.in_vals = []
    for key in new_dict.keys():
        new_query.in_cols.append(key)
        new_query.in_vals.append(str(new_dict[key]))
    new_query.build_str()

    db.conn.cursor().execute(new_query.build_insert())
    db.conn.commit()

    return


def edit(db, query):
    """ edit selection in query using provided up_vals and diplay change to user"""

    # copy query object but erase select columns to get all data for display
    query_all = copy.deepcopy(query)
    query_all.select_str = '*'

    # get pre-update data
    original_data = db.conn.cursor().execute(query_all.build_select()).fetchall()

    # update the data
    db.conn.cursor().execute(query.build_update())
    db.conn.commit()

    # get updated data
    updated_data = db.conn.cursor().execute(query_all.build_select()).fetchall()

    # display change to user
    print('Pre-Update:')
    # print(tabulate.tabulate(original_data, headers=header))
    show_tabulated_sql(db, query_all, data=original_data)
    print("Post Update:")
    # print(tabulate.tabulate(updated_data, headers=header))
    show_tabulated_sql(db, query, data=updated_data)


def tag_entry(db, tagged_query, tag_param):
    # TODO drop duplicates in tags_links instead of unique constraint
    """
    tag selections in tagged_query as tags in tag_query
    tagged query can be multiple items in selection but tag_query must only select single tage at a time

    args:
        db:                 FinanceDB object from finance_db lib
        tagged_query:       Query object with selection to be tagged
        tag_param:          tag_id or tag_desc to select tag to tag tagged_query selection
    """

    # create tags table if not exists
    if not db.exists('tags'):
        db.create_table('tags')

    # identify items to be tagged from transactions given user inputs

    # check if id column of data to be tagged is also in tags-links, otherwise table does not support tagging
    if db.schema[tagged_query.table][0] not in db.schema['tags_links']:
        print(" Table does not support tagging ")
        return

    # tag param can be either tag_id or tag_desc
    # tag query contains query to find tag to tag items passed in query or to create tag if not exists
    tags_table = 'tags'
    in_vals = [None, tag_param]
    w_cols = db.schema[tags_table]
    w_conds = ['=', '=']
    w_vals = [tag_param, tag_param]
    w_joins = ['OR', None]

    tag_query = finance_db.Query(table=tags_table, in_vals=in_vals,
                                 w_cols=w_cols, w_conds=w_conds, w_vals=w_vals, w_joins=w_joins)

    # need full data point to filtered columns and up_vals are cancelled out, only for when update and tag are
    # done simultaneously
    tagged_query.build_str()
    to_tag = db.conn.cursor().execute(tagged_query.build_select()).fetchall()

    # check if tag exists, create if not and get tag_data
    tag_data = db.conn.cursor().execute(tag_query.build_select()).fetchall()
    if len(tag_data) == 0:
        # tag does not exist
        create_item(db=db, query=tag_query, item_type='tag')
        tag_data = db.conn.cursor().execute(tag_query.build_select()).fetchall()[0]
        tag_id = str(tag_data[0])
    else:
        tag_id = str(tag_data[0][0])

    for transaction in to_tag:
        # tag transaction
        table = 'tags_links'
        query_tags_transactions = finance_db.Query(table=table, in_cols=db.schema[table],
                                                   in_vals=[str(transaction[0]), tag_id])
        db.conn.cursor().execute(query_tags_transactions.build_insert())

    show_tabulated_sql(db=db, query=tagged_query)
    print("Tagged above transactions with tags below!")
    show_tabulated_sql(db=db, query=tag_query)

    return


def categorize():
    BankClassify().ask_with_guess()



from docopt import docopt
import finance_db
import data
import docopt_utility
import copy

# sys.tracebacklimit = 0
# TODO revise usage to make actual sense
# TODO figure out how to limit inputs of arguments instead of parsing options into string after

usage = """

finance_py CLI.

Usage:
    data_driver.py create           ((--a <number> (--TD | --QT | --crypto) (--f | --api) [<description> <adjust>]) | 
                                    ( --t <tag_desc> )... | 
                                    ( --c <cat_desc> )... |
                                    ( --crh <holding_desc> <symbol> [( <parent_chain> <chain_addy> )]))                 
    data_driver.py view <table>     [(--s_col <s_cols>)...
                                    (--w <w_cols> <w_conds> <w_vals> [(--j <w_joins>)])...
                                    (--or <o_col> (--asc | --desc))
                                    (--l <limit>)]                             
    data_driver.py update           (--a (<account> | --all ))
    data_driver.py edit <table>     [(--u <up_cols> <up_vals>)... 
                                    (--w <w_cols> <w_conds> <w_vals> [(--j <w_joins>)])...
                                    (--t <tag_desc>...)]
    data_driver.py split            (--w <w_cols> <w_conds> <w_vals> [(--j <w_joins>)])...
                                    [[ (--p <percent>)  | (--n <new_value>) ] (--u <up_cols> <up_vals>...)]                               
    data_driver.py tag <table>      (--w <w_cols> <w_conds> <w_vals> [(--j <w_joins>)])...
                                    (--t <tag_desc>...)                                  
    data_driver.py delete <table>   [(--w <w_cols> <w_conds> <w_vals> [(--j <w_joins>)])...]
    data_driver.py process
    data_driver.py                  [--help]

Arguments:
    <account>               distinct account number or description
    <adjust>                starting value or adjustment if account data incomplete
    <cat_desc>              new category description
    <chain_addy>            chain/contract address for new crypto holding
    <description>           descriptive identifier or reference for new account
    <holding_desc>          new crypto holding description
    <parent_chain>          name of parent chain for new crypto holding e.g. ethereum for chainlink token
    <number>                new account number, wallet address if crypto
    <s_cols>                columns to be filtered in query/selection
    <symbol>                new crypto holding symbol, must match Yahoo ticker for accurate price information
    <tag_desc>              new tag description
    <w_joins>               keyword string to join where clause to the previous where clause, 'AND' or 'OR'
    <limit>

Options:
    --a                     create a new account  
    --all                   update all accounts
    --api                   new account will use api for updates
    --c                     new category
    --crh                   new crypto holding in blockchain wallet
    --crypto                new blockchain wallet
    --f                     new account will create folder and use files to update  
    --QT                    new Questrade account 
    --s_col                 add selected/filtered columns to query selection
    --t                     new tag
    --TD                    new TD account
    --u                     update accounts            
"""

args = docopt(usage)
print(args)

docopt_utility.elim_apostrophes(args=args)

# initialize objects to pass to various function calls
query = finance_db.Query(table=args['<table>'],
                         s_cols=args['<s_cols>'],
                         in_vals=None, in_cols=None,
                         up_vals=args['<up_vals>'], up_cols=args['<up_cols>'],
                         w_cols=args['<w_cols>'], w_conds=docopt_utility.process_clause(args), w_vals=args['<w_vals>'], w_joins=args['<w_joins>'],
                         o_col=args['<o_col>'], o_cond=docopt_utility.o_cond(args),
                         limit=args['<limit>'])

db = finance_db.FinanceDB()

if args['delete']:
    db.conn.cursor().execute(query.build_delete())
    db.conn.commit()

if args['process']:
    data.categorize()

if args['edit']:
    # add tags to in_vals and rebuild strings
    if args['<tag_desc>']:
        for tag in args['<tag_desc>']:
            data.tag_entry(db=db, tagged_query=query, tag_param=tag)

    data.edit(db=db, query=query)

if args['split']:
    query.table = 'transactions'
    data.split_transaction(db=db, query=query, percentage=args['<percent>'], amount=args['<new_value>'])

if args['create']:
    if args['--a']:
        # convert source option into string
        source = ""
        if args['--f']:
            source = 'file'
        elif args['--api']:
            source = 'api'

        # convert institution into string
        institution = ""
        if args['--TD']:
            institution = 'TD'
        elif args['--QT']:
            institution = 'QT'
        elif args['--crypto']:
            institution = 'crypto'

        query.table = 'accounts'

        query.in_vals = [None, args['<number>'], institution, args['<description>'], None, source, args['<adjust>']]
        query.build_str()

        comp_columns = copy.deepcopy(query.in_cols)
        comp_columns.pop(0)
        comp_columns.pop(-1)

        data.create_item(db=db, query=query, item_type='account', drop_cond='MIN', drop_method='inside',
                         comp_columns=comp_columns)

    if args['--t']:
        for tag in args['<tag_desc>']:
            query.table = 'tags'
            in_vals = [None, tag]
            query.in_vals = in_vals
            query.build_str()

            data.create_item(db=db, query=query, item_type='tag')

    if args['--c']:
        for category in args['<cat_desc>']:
            query.table = 'categories'
            in_vals = [None, category ]
            query.in_vals = in_vals
            query.build_str()

            data.create_item(db=db, query=query, item_type='category')

    if args['--crh']:
        query.table = 'crypto_holdings'
        in_vals = [None, args['<holding_desc>'], args['<symbol>'], args['<parent_chain>'], args['<chain_addy>']]
        query.in_vals = in_vals
        query.build_str()

        data.create_item(db=db, query=query, item_type='crypto_holding')

if args['tag']:
    for tag in args['<tag_desc>']:
        data.tag_entry(db=db, tagged_query=query, tag_param=tag)

if args['view']:
    if args['<table>'] in db.schema.keys():
        data.view(db=db, query=query)
    else:
        print("Table does not exist")

if args['--help']:
    print(usage)

if args['update']:
    if args['--a']:
        if args['<account>']:
            data.update_account(db=db, account=args['<account>'])

        elif args['--all']:
            accounts = db.conn.cursor().execute("SELECT num FROM accounts").fetchall()

            for account in accounts:

                data.update_account(db=db, account=account[0])



import sqlite3
import pathlib
import pandas as pd
import copy
import os

fipy_fp = pathlib.Path(__file__).absolute().parent
os.makedirs(fipy_fp.joinpath('db'), exist_ok=True)


class FinanceDB:
    """    Object for interacting with the finance database object

    Attributes:
    filename:   name of the database file including filepath - hardcoded
    filepath:   file location of the db file, based on parent file of src
    conn:       database connection object
    cursor:     database cursor object
    schema:     dictionary of lists, keys are tables, values are columns

    """

    def __init__(self):
        """ create database object and populate schema data"""
        # connect to db when initialized and establish cursor for later use
        self.filename = "CTFinance.db"
        self.filepath = fipy_fp.joinpath("db")

        # must enforce foreign keys when connection is formed
        self.conn = sqlite3.connect(self.filepath.joinpath(self.filename))
        self.conn.cursor().execute('PRAGMA foreign_keys = 1')

        # access and store schema
        self.schema = dict()
        self.schema = self._get_accounts()

    def refresh_conn(self):
        self.conn.close()
        self.conn = sqlite3.connect(self.filepath.joinpath(self.filename))
        self.conn.cursor().execute('PRAGMA foreign_keys = 1')

    def _get_accounts(self):
        tables = self.conn.cursor().execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()
        self.conn.cursor().close()
        schema = dict()
        # remove sqlite sequence table
        for tup in tables:
            if "sqlite_sequence" in tup:
                tables.remove(tup)
        # access and populate columns for each table name from db
        for table in tables:
            columns = list(map(lambda x: x[0], self.conn.cursor().execute("SELECT * FROM " + table[0]).description))
            schema[table[0]] = columns

        self.conn.cursor().close()
        return schema

    def _queryize_columns(self, columns_list):
        """create string compatible with injection in sql query from list of strings"""
        col_str = ""
        if columns_list:
            if len(columns_list) != 0:
                for column in columns_list:
                    if col_str == "":
                        col_str = col_str + column
                    else:
                        col_str = col_str + "," + column
        else:
            col_str = "*"

        return col_str

    def drop_duplicates(self, table, method, condition=None, new_data=None, id_col=None, filter_columns=None):
        """ Drop sql table entries based on first or last duplicate entered

        Args:
            <table>             string table name to delete duplicates
            <condition>         MIN or MAX, decides if maximum or minimum id in selected entries is not deleted
            <id_col>            which columns to use as unique id, default to first column in table schema
                                to this columns
            <filter_columns>    columns to determine if entry is unique or duplicate, if None all columns but first are
                                used as first is typically id column
        """
        if method in ['inside', 'outside']:
            # if no id_col passed, default is first column in table schema
            if not id_col:
                id_col = self.schema[table][0]

            # if not filter columns are passed than all are used except for the first [0] which is id, as all id are
            # unique and would delete nothing
            # copy.deepcopy(db.schema[query.table])
            if not filter_columns:
                columns_str = self._queryize_columns(copy.deepcopy(self.schema[table]).pop(0))
                filter_columns = copy.deepcopy(self.schema[table]).pop(0)
            else:
                columns_str = self._queryize_columns(filter_columns)

            if method == 'inside':
                if condition:
                    query = "DELETE FROM " + table + " WHERE " + id_col + " NOT IN (SELECT " + condition + \
                            "(" + id_col + ") FROM " + table + " GROUP BY " + columns_str + ")"
                    print(query)
                    self.conn.execute(query)
                    self.conn.commit()
                else:
                    print('Must provide condition when using inside method')

            elif method == 'outside':
                if not new_data.empty:
                    # select all existing table data using passed columns or default all
                    query = pd.read_sql_query("SELECT " + columns_str + " FROM " + table, self.conn)
                    existing_data = pd.DataFrame(query, columns=filter_columns)

                    # transactions table must be formatted as date to compare to new transactions data
                    if 'date' in existing_data.columns.tolist():
                        # convert to datetime
                        existing_data['date'] = pd.to_datetime(existing_data['date'])
                    if 'date' in new_data.columns.tolist():
                        new_data['date'] = pd.to_datetime(new_data['date'])

                    # merge data into single df
                    new_data = new_data.merge(existing_data, indicator=True, how='outer')
                    # merged dataframes will have _merge column indicating with dataframe contains each row
                    # drop rows where _merge column is 'both' eliminating overlap in files
                    new_data.drop(index=new_data[new_data['_merge'] == 'both'].index, inplace=True)
                    new_data.drop(index=new_data[new_data['_merge'] == 'right_only'].index, inplace=True)
                    # _merge column no longer needed
                    new_data.drop(columns=['_merge'], inplace=True)

                    # dump cleaned transactions to sql
                    new_data.to_sql(name=table, index=False, con=self.conn, if_exists='append')
                else:
                    print('Must provide new data to compare')

        else:
            print("Method must be either 'inside' or 'outside")

    def create_transactions(self,):

        # transactions has foreign keys linking to categories and accounts
        self.create_categories()
        self.create_accounts()

        transactions = """ CREATE TABLE IF NOT EXISTS transactions (
                            trans_id INTEGER PRIMARY KEY AUTOINCREMENT,
                            date TEXT NOT NULL,
                            desc TEXT NOT NULL,
                            amount REAL NOT NULL,
                            total_id REAL NOT NULL,
                            
                            processed INTEGER CHECK(processed IN (0,1)) DEFAULT(0),
                            
                            cat_id INTEGER DEFAULT NULL,
                            acc_id INTEGER,
                            FOREIGN KEY (cat_id) REFERENCES categories (cat_id) ON DELETE SET NULL ON UPDATE CASCADE,
                            FOREIGN KEY (acc_id) REFERENCES accounts (acc_id) ON DELETE CASCADE ON UPDATE CASCADE
                            
                        ); """

        # create slits to store originals of split transactions so they arent replicated in updates
        splits = """ CREATE TABLE IF NOT EXISTS splits (
                            trans_id INTEGER PRIMARY KEY AUTOINCREMENT,
                            date TEXT NOT NULL,
                            desc TEXT NOT NULL,
                            amount REAL NOT NULL,
                            total_id REAL NOT NULL
                            
                        ); """

        if 'transactions' not in self.schema.keys():
            self.conn.cursor().execute(transactions)
            print("Created transactions table")

        if 'splits' not in self.schema.keys():
            self.conn.cursor().execute(splits)
            print("Created splits table")

        # update schema!
        self.schema = self._get_accounts()

        return

    def create_categories(self,):
        categories = """ CREATE TABLE IF NOT EXISTS categories (
                            cat_id INTEGER PRIMARY KEY AUTOINCREMENT,
                            cat_desc TEXT NOT NULL
    
                        ); """

        if 'categories' not in self.schema.keys():
            self.conn.cursor().execute(categories)
            print("Created categories table")

        # update schema!
        self.schema = self._get_accounts()

        return

    def create_budgets(self,):
        budgets = """ CREATE TABLE IF NOT EXISTS budgets (
                        b_id INTEGER PRIMARY KEY AUTOINCREMENT,
                        b_resolution TEXT NOT NULL CHECK(b_resolution IN ('month','day')),
                        b_target REAL NOT NULL,
                        disp_start TEXT,
                        disp_resolution REAL TEXT CHECK(disp_resolution IN ('month', 'week', 'day')),
                        disp_end TEXT,
                        disp_window_size INTEGER,
                        disp_window_unit TEXT CHECK(disp_window_unit IN ('month', 'week', 'day')),
                        disp_type TEXT NOT NULL,
                        graph_type TEXT NOT NULL,
                        proj_type TEXT,
                        b_cat INTEGER NOT NULL,
                        FOREIGN KEY (b_cat) REFERENCES categories (cat_id) ON DELETE CASCADE ON UPDATE CASCADE
    
                        ); """

        if 'budgets' not in self.schema.keys():
            self.conn.cursor().execute(budgets)
            print("Created budgets table")

        # update schema!
        self.schema = self._get_accounts()

        return

    def create_dashboards(self,):
        dashboards = """ CREATE TABLE IF NOT EXISTS dashboards (
                                    dash_id INTEGER PRIMARY KEY AUTOINCREMENT,
                                    dash_desc TEXT NOT NULL,
                                    dash_date DATETIME DEFAULT CURRENT_TIMESTAMP
            
                                ); """

        if 'dashboards' not in self.schema.keys():
            self.conn.cursor().execute(dashboards)
            print("Created dashboards table")
        # update schema!
        self.schema = self._get_accounts()

    def create_budgets_dashboards(self,):
        budgets_dashboards = """ CREATE TABLE IF NOT EXISTS budgets_dashboardss (
                                    b_id INTEGER,
                                    dash_id INTEGER,
                                    disp_type TEXT NOT NULL,
                                    UNIQUE(b_id, dash_id)
                                    FOREIGN KEY (b_id) REFERENCES budgets (b_id) ON DELETE CASCADE ON UPDATE CASCADE,
                                    FOREIGN KEY (dash_id) REFERENCES dashboards (dash_id) ON DELETE CASCADE ON UPDATE CASCADE
            
                                ); """

        if 'budgets_dashboards' not in self.schema.keys():
            self.conn.cursor().execute(budgets_dashboards)
            print("Created budgets_dashboards link table")
        # update schema!
        self.schema = self._get_accounts()

    def create_accounts(self,):
        accounts = """ CREATE TABLE IF NOT EXISTS accounts (
                            acc_id INTEGER PRIMARY KEY AUTOINCREMENT,
                            num INTEGER NOT NULL UNIQUE,
                            institution TEXT,
                            desc TEXT,
                            filepath TEXT NOT NULL,
                            source TEXT CHECK(source IN ('file', 'api')),
                            adjust REAL
    
                        ); """

        if 'accounts' not in self.schema.keys():
            self.conn.cursor().execute(accounts)
            print("Created accounts table")
        # update schema!
        self.schema = self._get_accounts()

        return

    def create_crypto(self,):
        crypto = """ CREATE TABLE IF NOT EXISTS crypto (
                        cr_id INTEGER PRIMARY KEY AUTOINCREMENT,
                        date DATETIME DEFAULT CURRENT_TIMESTAMP,
                        desc TEXT NOT NULL,
                        qty REAL NOT NULL,
                        price REAL NOT NULL,
                        total REAL NOT NULL
                        
                    ); """

        if 'crypto' not in self.schema.keys():
            self.conn.cursor().execute(crypto)
            print("Created crypto table")
        # update schema!
        self.schema = self._get_accounts()

        return

    def create_crypto_holdings(self,):
        crypto_holdings = """ CREATE TABLE IF NOT EXISTS crypto_holdings (
                        crh_id INTEGER PRIMARY KEY AUTOINCREMENT,
                        desc TEXT NOT NULL,
                        symbol TEXT NOT NULL,
                        parent_chain TEXT,
                        chain_address TEXT
                        
                    ); """

        if 'crypto_holdings' not in self.schema.keys():
            self.conn.cursor().execute(crypto_holdings)
            print("Created crypto_holdings table")
        # update schema!
        self.schema = self._get_accounts()

        return

    def create_qtrade(self,):
        qtrade = """ CREATE TABLE IF NOT EXISTS qtrade (
                        qt_id INTEGER PRIMARY KEY AUTOINCREMENT,
                        date DATETIME DEFAULT CURRENT_TIMESTAMP,
                        symbol TEXT NOT NULL,
                        symbolId TEXT NOT NULL,
                        openQuantity INTEGER,
                        closeQuantity INTEGER,
                        currentMarketValue REAL NOT NULL,
                        currentPrice REAL NOT NULL,
                        averageEntryPrice REAL NOT NULL,
                        totalCost REAL NOT NULL
                        
                    ); """

        if 'qtrade' not in self.schema.keys():
            self.conn.cursor().execute(qtrade)
            print("Created qtrade table")
        # update schema!
        self.schema = self._get_accounts()

        return

    def create_tags(self,):
        self.create_transactions()
        tags = """ CREATE TABLE IF NOT EXISTS tags (
                            tag_id INTEGER PRIMARY KEY AUTOINCREMENT,
                            desc TEXT NOT NULL UNIQUE
    
                        ); """

        tags_links = """ CREATE TABLE IF NOT EXISTS tags_links (
                            trans_id INTEGER,
                            tag_id INTEGER,
                            UNIQUE(trans_id, tag_id)
                            FOREIGN KEY (trans_id) REFERENCES transactions (trans_id) ON DELETE CASCADE ON UPDATE CASCADE,
                            FOREIGN KEY (tag_id) REFERENCES tags (tag_id) ON DELETE CASCADE ON UPDATE CASCADE
    
                        ); """

        if 'tags' not in self.schema.keys():
            self.conn.cursor().execute(tags)
            self.conn.cursor().execute(tags_links)
            print("Created tags table")

        # update schema!
        self.schema = self._get_accounts()

        return

    def create_table(self, name):
        """ create default tables used in finance app using default schema set internally
        Args:
        name:   name of table to create(transactions, accounts, categories)

        """

        if name not in self.schema.keys():
            if name == 'transactions':
                self.create_transactions()

            elif name == 'categories':
                self.create_categories()

            elif name == 'accounts':
                self.create_accounts()

            elif name == 'tags':
                self.create_tags()

            elif name == 'crypto':
                self.create_crypto()

            elif name == 'crypto_holdings':
                self.create_crypto_holdings()

            elif name == 'qtrade':
                self.create_qtrade()

            elif name == 'budgets':
                self.create_budgets()
                self.create_dashboards()
                self.create_budgets_dashboards()

            elif name == 'dashboards':
                self.create_budgets()
                self.create_dashboards()
                self.create_budgets_dashboards()

            else:
                pass
        else:
            pass

        return

    def exists(self, table):
        """ Check if table exists"""
        if table in self.schema.keys():
            return True
        else:
            return False


db = FinanceDB()


class Query:
    """ Store query parameters and create queries

        args:
            table:          str of table to perform query operations within
            s_cols:         list of column names as str to be queried
            in_vals:        list of values to be inserted
            up_vals:        list of values to be updated
            w_cols:         list of columns as string to construct query where clauses
            w_conds:        list of operators as strings to use in constructing query where clauses
            w_vals:         list of values to construct where clauses e.g. 'WHERE w_col[1] w_cond[1] w_val[1]'
            o_col:          string column name to be used in order clause
            o_cond:         string denoting order condition, can be 'ASC' or 'DESC', e.g. 'ORDER BY o_col o_cond'
            limit:          string number denoting limit clause, e.g. 'LIMIT limit'
            build_str:      bool indicating if all str properties should be built @ init or later if false
                            allows for addition/modification of query properties post init

        props:
            where_str:      query ready string containing where clauses
            select_str:     query ready string containing select columns
            order_str:      query ready string containing order clause
            limit_str:      query ready string containing limit clause
            values_str:     query ready string containing values for insert
            update_str:     query ready string containing cols/vals for updating
            table:          string table name in which queries operate
            wheres:         list of dictionaries structured [{'col': w_col, 'cond': w_cond, 'va': w_val}]
            o_col:          string column name to order query by
            o_cond:         string condition to order query by, 'ASC' or 'DESC'
            in_vals:        list of values to be inserted - must correspond in order to s_cols
            in_cols:        corresponding list of columns as strings to in_vals
            up_vals:        list of values to be updated - must correspond in order to s_cols
            up_cols:        corresponding list of columns as strings to up_vals
            limit:          string value indicating limit or return query items

        """

    def __init__(self, table, s_cols=None,
                 in_vals=None, in_cols=None,
                 up_vals=None, up_cols=None,
                 w_cols=None, w_conds=None, w_vals=None, w_joins=None,
                 o_col=None, o_cond=None,
                 limit=None,
                 build_str=True):

        # self.db = FinanceDB()
        # initialize properties
        self.table = table
        self.select_str = ""
        self.in_values_str = ""
        self.in_cols_str = ""
        self.update_str = ""
        self.where_str = ""
        self.order_str = ""
        self.limit_str = ""
        self.wheres = []

        if not db.exists(self.table):
            db.create_table(self.table)

        self.w_joins = w_joins
        self.o_col = o_col
        self.o_cond = o_cond
        self.in_vals = in_vals
        self.in_cols = in_cols
        self.up_vals = up_vals
        self.up_cols = up_cols
        self.limit = limit

        # if no s_cols passed then use * (all in sql)
        if s_cols:
            self.s_cols = s_cols
        else:
            # default to all columns instead of *
            self.s_cols = '* '

        if in_vals:
            self.in_vals = self.replace_null(self.in_vals)
            if not in_cols:
                # default to all columns for insert if none passed and in_vals passed
                self.in_cols = db.schema[self.table]

        if up_vals and not up_cols:
            # default to all columns for update if none passed with up_vals
            self.up_cols = db.schema[self.table]

        if w_cols:
            for i in range(0, len(w_cols)):
                where = {}
                # sql will only recognize dates enclosed in apostrophes
                if 'date' in w_cols[i]:
                    w_vals[i] = "'" + w_vals[i] + "'"

                # process all else normally
                where['col'] = w_cols[i]
                where['cond'] = w_conds[i]
                where['val'] = w_vals[i]
                self.wheres.append(where)

        # as long as build str is not false, all query strings will be built
        if build_str:
            self.build_str()

    def build_str(self):
        """ build all query strings given Query properties """
        if not db.exists(self.table):
            db.create_table(self.table)

        # if * in s_cols, select_str is simply * as well
        if '*' not in self.s_cols:
            self.select_str = self._queryize_columns(select=True)
        else:
            self.select_str = '*'

        if self.in_vals:
            self.in_vals = self.replace_null(self.in_vals)
            if not self.in_cols:
                # default to all columns for insert if none passed and in_vals passed
                self.in_cols = db.schema[self.table]
            self.in_values_str = self._queryize_values()
            self.in_cols_str = self._queryize_columns(insert=True)

        if self.up_vals:
            self.update_str = self._queryize_updates(columns=self.up_cols, values=self.up_vals)

        if self.wheres:
            self.where_str = self._queryize_where()

        if self.o_col:
            self.order_str = "ORDER BY " + self.o_col + ' ' + self.o_cond + ' '

        if self.limit:
            self.limit_str = "LIMIT " + self.limit + ' '

        return

    def replace_null(self, p_list):
        """ replace NoneType objects with NULL"""
        p_list = ['NULL' if v is None else v for v in p_list]
        return p_list

    def _queryize_columns(self, select=None, insert=None):
        """create string compatible with injection in sql query from list of strings"""
        col_str = ""
        param = None
        if select:
            param = self.s_cols
        elif insert:
            param = self.in_cols

        if len(param) > 1:
            for col in param:
                if col_str == "":
                    col_str = col_str + col
                else:
                    col_str = col_str + "," + col
        else:
            col_str = param[0]

        return col_str

    def _queryize_values(self,):
        """create string compatible with injection in sql query from list of strings"""
        val_str = ""
        param = self.in_vals

        if len(param) > 1:
            for val in param:
                # replace val with param if injection option not chosen
                if not isinstance(val, str):
                    val = str(val)
                    val_str = val_str + val
                    # remove apostrophe as sql thinks its a string
                    val_str = val_str[0:(-1*(len(val))):] + val_str[(-1*(len(val)))::]
                    val_str += ","

                elif val == 'NULL':
                    val_str = val_str + val + ","
                else:
                    val_str = val_str + "'" + val + "',"
            val_str = val_str[:-1]
        else:
            val_str = param[0]



        return val_str

    def _queryize_where(self):
        """ create queryable string from where properties: WHERE col cond val (AND...) """

        where_str = "WHERE "

        # need to wrap all vals in aprotrophe or might nto work if not already for fucks sake
        for val in self.wheres:
            if val['val'][0] != "'" and val['val'][-1] != "'":
                val['val'] = "'" + val['val'] + "'"

        if len(self.wheres) > 1:
            # w_joins counter
            j = 0
            for i in range(0, len(self.wheres)):
                # if not last where condition then add w_join to prep for next statement
                # w_join is added prior to the where clause, do not add ot first clause
                if i != 0:
                    where_str = where_str + " " + self.w_joins[j] + " "
                    j = j + 1
                else:
                    where_str = where_str + " "

                # like conditions must be wrapped in ''
                if self.wheres[i]['cond'] == 'LIKE':
                    self.wheres[i]['val'] = "'" + self.wheres[i]['val'] + "'"

                where_str = where_str + self.wheres[i]['col'] + ' ' + self.wheres[i]['cond'] + ' ' \
                            + self.wheres[i]['val']

        else:
            if self.wheres[0]['cond'] == 'LIKE':
                self.wheres[0]['val'] = "'" + self.wheres[0]['val'] + "'"
            where_str = where_str + self.wheres[0]['col'] + ' ' + self.wheres[0]['cond'] + ' ' + self.wheres[0]['val']

        return where_str

    def _queryize_updates(self, columns, values):
        """create string compatible with injection in sql query from list of strings"""
        updates_str = ""
        # if string is used as a value it must be wrapped in '
        if isinstance(values, list):

            for i in range(0, len(columns)):
                if updates_str == "":
                    updates_str = columns[i] + " = '" + str(values[i]) + "'"
                else:
                    updates_str = updates_str + ", " + columns[i] + " = '" + str(values[i]) + "'"
        else:
            if isinstance(values, str):
                if values[0] != "'" and values[len(values)] != "'":
                    values = "'" + values + "'"
            updates_str = columns[0] + " = " + values

        return updates_str

    def build_select(self, function=None):
        # SELECT s_cols FROM table WHERE ORDER LIMIT
        # SELECT function(s_cols) FROM table WHERE ORDER LIMIT

        # data validation for functions and select query argument requirements
        available_functions = ['AVG', 'SUM']
        if function:
            if function not in available_functions:
                print('Function not included, available functions must be one of ' + str(available_functions)[1:-1])
                return
            if len(self.s_cols) != 1 or '*' in self.s_cols:
                print("Function can only be used with single column")
                return

        # always space after latest addition to query - standardize
        query = 'SELECT '

        # build query
        if function:
            query = query + function + '(' + self.select_str + ') '
        elif self.s_cols:
            query = query + self.select_str

        query = query + ' FROM ' + self.table + ' '

        if self.wheres:
            query = query + self.where_str

        if self.o_col:
            query = query + self.order_str

        if self.limit:
            query = query + self.limit_str

        print(query)
        return query

    def build_insert(self, method='FAIL',):
        # INSERT OR method INTO table (columns,) VALUES (values,)
        query = 'INSERT OR ' + method + ' INTO ' + self.table + \
                ' (' + self.in_cols_str + ') ' + 'VALUES ' + '(' + self.in_values_str + ')'

        print(query)
        return query

    def build_update(self, method='FAIL'):
        # INSERT OR method INTO table (columns,) VALUES (values,)
        query = 'UPDATE OR ' + method + ' ' + self.table + ' SET ' + self.update_str + ' '

        if self.wheres:
            query = query + self.where_str

        print(query)
        return query

    def build_delete(self):
        # TODO DROP TABLE
        # DELETE s_cols FROM table WHERE ORDER LIMIT

        # build query
        query = 'DELETE FROM ' + self.table + ' '

        if self.wheres:
            query = query + self.where_str

        print(query)
        return query



from distutils.core import setup

setup(name='caretaker_fdata',
      version='1.0',
      packages=['src'],
      # Start with a small number and increase it with every change you make
      license='MIT',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
      description = 'finance management tool',   # Give a short description about your library
      author = 'Dan',                   # Type in your name
      author_email = 'daniel.js.campbell@gmail.com',      # Type in your E-Mail
      url = 'https://github.com/dn757657/caretaker_f2',   # Provide either the link to your github or to your website
      download_url = 'https://github.com/dn757657/ct_data/archive/refs/tags/test.tar.gz',    # I explain this later on
      keywords = ['Management', 'finance', 'automation'],   # Keywords that define your package best
      install_requires=[            # I get to this in a second
            'pandas',
            'tabulate',
            'web3',
            'python-dateutil',
            'textblob',
            'colorama',
            'docopt',
            'qtrade',
            'pandas_datareader',
            'textblob',
      ],
      classifiers=[
            'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
            'Intended Audience :: Developers',      # Define that your audience are developers
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',   # Again, pick a license
            'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
      ],
      )



from qtrade import Questrade
import pathlib
import finance_db

fipy_fp = pathlib.Path(__file__).absolute().parent.parent
src_path = fipy_fp.joinpath('src')
yaml_path = src_path.joinpath('access_token.yml')

def qtrade_connect():
    print(yaml_path)
    try:
        # access_code = 'hJxEWwStLWDPAx397CYGXc93AtQEeAa10'
        # qtrade = Questrade(access_code=access_code)
        qtrade = Questrade(token_yaml=yaml_path)
        qtrade.refresh_access_token(from_yaml=True)
    except:
        # qtrade = Questrade(token_yaml=yaml_path)
        # qtrade.refresh_access_token(from_yaml=True)
        access_code = input("please input Questrade API Access token ")
        qtrade = Questrade(access_code=access_code)

    return qtrade


def update_qpositions(db, account_id):
    if not db.exists('qtrade'):
        db.create_table('qtrade')

    qtrade = qtrade_connect()
    positions = qtrade.get_account_positions(account_id=account_id)
    for pos in positions:
        update = {key: None for key in db.schema['qtrade']}
        for key in update.keys():
            # db table column names are same as default dictionary names for properties in position
            # need to eliminate params that are in positions dictionaries but not in db qtrade table
            if key in pos.keys():
                update[key] = pos[key]

        update.pop('date')
        update_cols = list(update.keys())
        update_vals = list(update.values())

        query = finance_db.Query(table='qtrade',
                                 in_vals=update_vals, in_cols=update_cols,)

        # query = db.insert(table='qtrade', columns=update_cols, values=update_vals)
        db.conn.cursor().execute(query.build_insert())
        db.conn.commit()
